/*-------------------------------------
  * fichier:        main.c
  * objet:          programme principal gestion horloge DCF77
  * Auteur:         Pascal Chour
  * Site            www.pascalchour.fr
  * date            22/08/2023
  * Version         1.0
  *
  * Ce programme est une simple horloge qui ne fait rien d'autre qu'afficher l'heure et la date.
  * Sa mise  l'heure est faite automatiquement par un module DCF77. Entre deux mise  l'heure, elle fonctionne en roue libre (sur son propre quartz).
  * L'horloge tente une mise  l'heure toute les 24H00  00:00:00.
  * Il est possible optionellement de forcer la mise  l'heure par un bouton poussoir.
  * L'horloge affiche l'heure (HH:MN:SS) et alternativement, l'anne et le jour/mois
  * L'indicateur de rception est allum lorsque l'horloge tente de lire le signal DCF77.
  * L'indicateur (cl 6 pans = outil) est affich lorsque l'horloge n'a pas pu lire le signal DCF77 depuis plus de 24 heures.
  *
  * Les GPIO sont programms en dur.
  * P20 = entre signal DCF77
  * P21 = signal de sortie d'activation du module DCF77 (1 = activ, 0 = dsactiv)
  * P3x2 = entres bouton de forcage de mise  l'heure (ceux de la carte Olimex).
  *
  * Le fonctionnement normal est le suivant :
  * En fonctionnement sans DCF77 actif, les boutons sont actifs, le module DCF77 est dsactiv et l'interruption sur le signal (P20) est dsactiv.
  * Lorsque le module DCF77 est activ (P21), les boutons sont dsactivs et l'interruption sur le signal est active (P20).
  * lorsqu'on appuie sur un bouton, on active le module DCF77 (P21) et l'interruption sur le signal (P20) et on dsactive les boutons.
  *
  * La recherche du signal se fait pendant 10 minutes maximum. Si le signal n'a pas t reu correctement en 10mn, le DCF77 est dsactiv et on signal une erreur en faisant clignoter l'afficheur.
  * la recherche du signal est dclenche automatiquement  00:00:00 ou lorsqu'on appuie sur le bouton.
  *
  * Le processeur fonctionne  1MHz. L'horloge basse frquence 32768Hz est utilise pour la gestion de l'horloge en roue libre (IT toutes les secondes)

*/

#include <msp430x44x.h>
#include <stdio.h>
#include <stdint.h>
#include "lcd.h"
#include "dcf77util.h"
#include "common.h"

/* Variables d'tat du systme (globales mais on peut mettre la plupart dans le main) */

  enum T_ETATSIGNAL {PREMIER0, TRAITE0, TRAITE1, TERMINE};
  enum T_ETATBIT {TA_MESURE_SIGNAL, TA_MESURE_SYNCHRO, TA_MISE_A_HEURE, TA_FIN};

  const char s_efface[]  = "      ";


  unsigned int HH;            // Heure courante
  unsigned int MN;            // Minutes courantes
  unsigned int SS;            // Secondes courantes
  unsigned int Annee;         // Anne
  unsigned int Mois;          // Mois
  unsigned int Jour;          // Jour
  unsigned int HHtemp, MNtemp, JJtemp, MMtemp, AAtemp; // pour dcodage signal horaire, variables temporaires
  int TO_Trame ;              // Time Out Trame
  int TO_Signal;              // Time out sur signal


  bool_t reception_active;    // vrai si rception en cours
  bool_t trame_recue;         // Varai si on a reu une trame complte. La date et l'heure ont t mises  jour.
  bool_t bascule_1s;          // change d'tat toutes les secondes.
  bool_t alarme_mise_a_heure; // Vrai si on n'arrive pas  mettre l'horloge  l'heure avec le module
  bool_t ITWatchdog;          // Une IT du Wtachdog  traiter (toute les secondes)
  bool_t ITMilliseconde;      // Une IT d'un timer  traiter toutes les millisecondes
  bool_t ITFront;             // Une IT sur front montant ou descendant  traiter
  uint32_t TCourantFront;     // Timer du dernier front
  uint32_t TCourantMS;        // Timer de la dernire IT en millisecondes
  unsigned long TimerChangementEtat;   // Timer de la dernire mesure de transition sur P2.0
  bool_t Stocke = FALSE;      // Sert  l'automate de traitement des bits pour savoir s'il doit stocker le rsultat dans le buffer de rception
  unsigned int IBufRecep;     // Index courant dans le Buffer de rception
  int BufRecep[60];           // Buffer de rception 60=>Le signal fait 60 bits

  uint32_t DebutMesure;       // Dbut de la mesure d'une nouvelle trame
  int ValeurMesure;
  int DureeSignal;
  unsigned int NombreParasites; // Pour compter le nombre de parasites par seconde

  enum T_ETATSIGNAL ETATSIGNAL;     // Etat automate traitement du signal
  enum T_ETATBIT ETATBIT;           // Etat automate traitement des bits


  /*
  Cette fonction active ou dsactive le module DCF77.
  on = TRUE : rception active
  on = FALSE : rception inactive

  Lorsqu'on active la rception, la fonction vrifie si elle n'est pas dj active afin de ne pas initialiser de nouveau certaines choses.
  Lorsque la rception est inactive, la fonction :
      - initialise un indicateur pour signifier que la trame n'a pas t reue
      - active le module DCF77
      - active les timers
      - affiche l'antenne (sans le bargraph)
      - active les IT sur fronts pour rcuprer le signal du module DCF77 (on attend un front bas pour commencer)
      - indique que la rception est active
      - initialise les tats des deux automates : attente d'un front bas (PREMIER0) et mesure du signal de synchro (TA_MESURE_SYNCHRO)
      - initialise les timers pour les timeout.

  Lorsqu'on dsactive la rception, la fonction:
      - dsactive les timers
      - dsactive les IT sur fronts montans ou descendants
      - dsactive le module DCF77
      - supprime l'affiche de l'antenne (et du bargraph)

  */

  void ActiveDCF77(bool_t on)
  {
      if (on) {
          if (!reception_active) {        // si rception dj active, on se contente de rinitialiser le time out
              trame_recue = FALSE;        // On n'a pas encore reu la trame
              P2OUT &= 0xFD;              // DCF77 activ
              LCD_antenne(TRUE);          // On indique que l'on recherche le signal.
              ETATSIGNAL = PREMIER0;      // Attente d'un signal  0
              ETATBIT = TA_MESURE_SYNCHRO; // Mesure de la synchronisation
              Timer1ms(TRUE);             // On active le timer 1ms
              P2IES = P2IES |= 0x01;      // On attend un front bas pour se synchroniser sur un signal  0
              P2IE |= 0x01;               // IT sur signal DCF77 activ (P2.0)
              NombreParasites = 0;        // A priori, pas (encore) de parasites
              reception_active = TRUE;    // On indique que la rception est active
          }
          TO_Trame = Attente_Trame_Max;   // pour time out attente trame
          TO_Signal = 0;                  // pour time out attente signal
      } else {
          P2IE &= 0xFE;                   // IT DCF77 dsactiv
          P2OUT |= 0x02;                  // DCF77 dsactiv
          Timer1ms(FALSE);                // On dsactive le timer 1ms
          LCD_antenne(FALSE);             // On efface l'antenne et le bargraph
          reception_active = FALSE;       // On indique que la rception n'est plus active
      }
  }


/**
 * Programme principal. Iniialisation et gestion boucle horloge
 */
int main (void)
{
    char s[20];

    // INITIALISATIONS //
    WDTCTL = WDTPW | WDTHOLD;                       // Password + stop watchdog timer

    FLL_CTL0 &= ~XTS_FLL;                           // XT1 as low-frequency
    // !!!!!!!!!!!
    FLL_CTL0 = XCAP10PF;                            // set load capacitance for 32k xtal (18PF dans la version exemple de Peter Ivanov)
    // !!!!!!!!!!!

    _BIC_SR (OSCOFF);                               // Mise en route de l'oscillateur

    do                                              // Attente jusqu' ce que le signal d'horloge soit stable
    {
        IFG1 &= ~OFIFG;
    }
    while (IFG1 & OFIFG);





    FLL_CTL1 = 0x60;                                // Initialisation de CTL1 : SMCLK OFF, XT2 OFF, MCLK = DCO, SMCLK = DCO, ALCK/1.

    // Initialisation des ports. A la mise sous tension, ils sont en entre. Aprs quelques tests, la configuration suivante est celle qui
    // semble le moins consommer : P1 en sortie, P2 en sortie (sauf DALLAS et "DCF77 enabled"). Les autres ports en entre
    // sachant que beaucoup d'entre-eux (P5, P6) sont initialiss par le contrleur LCD.
    P1DIR = 0xFF;
    P1OUT |= STATUS_LED;                            // STATUS LED teint

    P2DIR = 0x7E;
    P2OUT |= DCF77_Disabled;                        // DCF77 dsactiv

    LCD_init ();
    LCD_clear ();                                   // Efface l'afficheur


/* POUR TEST
    LCD_test();                                     // Test chaque segment. A utiliser avec debogueur
    LCD_clear ();                                   // Efface l'afficheur aprs le test
FIN TEST */

    setStatusLed (TRUE);                            // Allume diode
    LCD_write("CHOUR ",6);                          // Affiche n de version et autres
    delayMs(1000);
    sprintf(s,"%s%s","P ",version_pgm); 			// 4 caractres max
    LCD_write(s,7);
    delayMs(1000);
    setStatusLed (FALSE);

    // Initialisation de quelques variables d'tat
    bascule_1s = FALSE;             // il faut lui donner une valeur...
    alarme_mise_a_heure = TRUE;     // Au dmarrage, on signale que l'horloge n'est pas synchronise
    ITWatchdog = FALSE;             // Pas d'IT du timer
    trame_recue = FALSE;            // Pas de trame reue
    reception_active = FALSE;       // Pas de rception en cours
    HH = 23;                        // Heure choisie pour forcer un affichage complet de l'heure au dmarrage et pour lancer une mise  l'heure par le module DCF77
    MN = 59;
    SS = 59;
    Annee = MILLENAIRE+23;          // Date minimum pour ce programme (22 aot 2023)
    Mois = 8;
    Jour = 22;
    LCD_deuxpoints (1, TRUE);       // Affichage deux points sur afficheur bas (sparations heures:minutes:secondes)
    LCD_deuxpoints (2, TRUE);

    // Initialisation du WatchDog
    WDTCTL = WDT_ADLY_1000;         // WDT interruption toute les secondes (pilot par XT1)
    IE1 |= WDTIE;                   // Active IT WDT
    _EINT ();

    while (1) {


        // Traitement IT sur front montant ou descendant
        if (ITFront) {
            if ((!trame_recue) && reception_active) { // Si rception inactive, pas de traitement. Si Trame reue, pas de traitement
                TO_Signal = Attente_Signal_Max; // On remet  jour le timeout sur Signal

                //if (P2IES & 0x01) {setStatusLed (FALSE);} else {setStatusLed (TRUE);} // pour test

                if (DureeVraie(TimerChangementEtat, TCourantFront) > DureeParasite) { // Si dure de l'impulsion est infrieure  la dure d'un parasite, on n'en tient pas compte

                    switch (ETATSIGNAL) { // Recherche du signal de synchro

                        case PREMIER0: // Attente du premier signal  0 pour une synchronisation
                            if (P2IES & 0x01) { // Si on passe  l'tat bas (interruption sur front descendant)
                                ValeurMesure = 0;
                                DureeSignal = 0;
                                ETATSIGNAL = TRAITE0;
                                ETATBIT = TA_MESURE_SYNCHRO;
                            }
                        break;

                        case TRAITE0: // Traitement du signal  0. On arrive ici parce qu'on attendait un front montant
                            if ((DureeSignal > 1700) && (ValeurMesure < 0)) { // Si Dure signal  0 > 1700 et la mesure est un 0, on a trouv la synchronisation qu'on attendait
                                ETATSIGNAL = TERMINE; // Etat d'attente o on ne fait rien
                                IBufRecep = 0; // On dbute la capture de la trame
                                DebutMesure = TCourantFront; // Temps du dbut de la trame
                                ValeurMesure = 0;
                                ETATBIT = TA_MESURE_SIGNAL;
                            } else {
                                ETATSIGNAL = PREMIER0;
                            }
                        break;

                        case TERMINE:
                        break;


                    } // Fin Switch ETATSIGNAL
                } else { // Parasites. On comptabilise. L'affichage et la mise  jour des compteurs est gr par le module sous timer A0.
                    NombreParasites += 1;
                }
                P2IES ^= 0x01; // On change l'tat du front attendu
                TimerChangementEtat = TCourantFront; // Sert  la mesure de la largeur des impulsions
            } // Fin Trame reue. Pas de traitement
            ITFront = FALSE;
        }

        // Timer sur millisecondes
        if (ITMilliseconde) {
            uint32_t Diff;


            switch (ETATBIT) {
                case TA_MESURE_SYNCHRO:
                    if (P2IN & 0x01) {ValeurMesure++;} else {ValeurMesure--;}
                    DureeSignal++;
                break;
                case TA_MESURE_SIGNAL: // Programmation pour viter une suite de if then else
                    Diff =  DureeVraie(DebutMesure,TCourantMS) % 1000;
                    if (Diff < TDeb) { // < 130ms : on ne mesure pas encore le signal
                        Stocke = FALSE; // mais on positionne un indicateur pour la suite
                        break;
                    }
                    if (Diff < TFin) { // entre 130ms et 170ms, on chantillonne le signal
                        if (P2IN & 0x01) {ValeurMesure++;} else {ValeurMesure--;}
                        break;
                    }
                    if (!Stocke) { // Si on n'a pas encore stock la valeur chantillone, alors on le fait
                        Stocke = TRUE; // Pour ne pas restocker.
                        if (ValeurMesure >=0) {BufRecep[IBufRecep] = 1;} else {BufRecep[IBufRecep] = 0;} // ValeurMesure > 0 : on a majoritairement un signal haut => Valeur 1. Sinon, 0
                        if (((IBufRecep == 0) && (BufRecep[0] == 1)) || ((IBufRecep == 20) && (BufRecep[20] == 0))) { // bits 0 et 20 incorrects. Pas bon !
                            ETATBIT = TA_MESURE_SYNCHRO;
                            ETATSIGNAL = PREMIER0;
                            break;
                        }
                        IBufRecep++; // On incrmente l'index du stockage
                        if (IBufRecep >= 59) {
                            if (!VerifieBufRecep(BufRecep)) {
                                ETATBIT = TA_MESURE_SYNCHRO;
                                ETATSIGNAL = PREMIER0;
                            } else {
                                ETATBIT = TA_MISE_A_HEURE;
                                IE1 ^= WDTIE;                // Dsactive IT Watchdog et on laisse  la routine principale le temps de se terminer
                            }
                        } else {
                            ValeurMesure = 0;
                        }
                    }
                break;

                // L'ide est qu'aprs la vrification de la rception (avec rsultat OK), on se trouve dans la 59me seconde du signal.
                // On dsactive l'IT su Watchdog. Si le programme principal tait en cours d'excution, il se termine et boucle ensuite dans
                // le vide car IT_Timer est toujours  faux (plus d'IT du Watchdog).
                // Dans cet tat, on attend d'tre  la fin de la 59me seconde (environ 59999ms) puis on initialise la nouvelle heure/date que l'on affiche et enfin,
                // on remet en route le timer du Watchdog tout en le rinitialisant  0.
                // La prochaine IT se produira donc  la 61me seconde du dnut de la rception de la synchronisation et incrmentera le compteur des secondes
                // qui a t forc  zro lors de la remise  l'heure. On a donc une incertitude de quelques millisecondes sur les secondes qui dpend de la frquence
                // d'chantillonnage et du temps d'excution du programme.
                case TA_MISE_A_HEURE:
                    if (DureeVraie(DebutMesure,TCourantMS)  >= 59999) {
                        HH = HHtemp;
                        MN = MNtemp;
                        SS = 0;
                        Jour = JJtemp;
                        Mois = MMtemp;
                        Annee = MILLENAIRE+AAtemp;
                        trame_recue = TRUE;
                        SS = 0;
                        LCD_2digitafficheurbas(4,HH);
                        LCD_2digitafficheurbas(2,MN);
                        LCD_2digitafficheurbas(0,SS);
                        // Rinitialisation du WatchDog
                        WDTCTL = WDT_ADLY_1000;                 // WDT interruption toute les secondes, remise  zro du compteur
                        IE1 |= WDTIE;                           // Active IT Watchdog
                        ITWatchdog = FALSE;                     // au cas o...
                        ETATBIT = TA_FIN;                       // Le traitement est termin.
                    }
                break;

                case TA_FIN:
                break;
            }
            ITMilliseconde = FALSE;
        }

        // Timer sur watchdog (toute les secondes)
        // En rgime normal, on incrmente les secondes chaque seconde ! Mais lorsqu'une mise  l'heure vient d'tre faite
        //  partir du module DCF77, il ne faut pas incrmenter les secondes tout de suite. Pour cette raison, le programme qui met l'heure  jour
        // rinitialise le Watchdog (compteur = 0) et supprime l'ventuelle ITWatchdog en attente. La prochaine ractivation de cette partie du
        // programme n'aura donc lieu qu'une seconde aprs cette mise  l'heure.
        if (ITWatchdog) {
            bascule_1s = !bascule_1s; // Pour crer des alternances et des clignotements toutes les secondes
            TO_Trame--; // On dcrmente le timer pour le time out sur attente trame. Pas de risque de dpassement
            TO_Signal--; // On dcrmente le timer pour le time out sur attente signal. Pas de risque de dpassement

            // Si alarme mise  l'heure, on fait un signalement au niveau de l'affichage
            if (bascule_1s) {
                LCD_4digitafficheurhaut (Annee, FALSE);
                if (alarme_mise_a_heure) LCD_cle(TRUE);
            } else {
                LCD_4digitafficheurhaut (Jour*100+Mois, TRUE);
                if (alarme_mise_a_heure) LCD_cle(FALSE);
            }

            // Si rception active, on vrifie si on a reu quelque chose avant le time out
            if (reception_active) {
                if (trame_recue) {
                    ActiveDCF77(FALSE);             // On arrte la rception
                    alarme_mise_a_heure = FALSE;    // Plus d'alarme
                    LCD_cle(FALSE);                 // On efface le symbole comme quoi il y a des problmes
                } else {
                    if (TO_Trame <= 0) {
                        ActiveDCF77(FALSE);         // On arrte la rception
                        alarme_mise_a_heure = TRUE;
                        LCD_effaceafficheurhaut();  // On efface la date qui n'est plus crdible
                    } else {
                        if (TO_Signal <= 0) {
                            LCD_antenne(TRUE);      // On efface le bargraph
                        } else {
                            GereBargraph(&NombreParasites); // Pour affichage du Bargraph
                        }
                    }
                }
            } else { // Si rception inactive, alors on teste les boutons
                if (buttonHandler()) ActiveDCF77(TRUE);
            }


            SS = SS+1;
            if (SS >= 60) {
                SS = 0;
                LCD_2digitafficheurbas(0,SS);
                MN = MN+1;
                if (MN >= 60) {
                    MN = 0;
                    LCD_2digitafficheurbas(2,MN);
                    HH = HH+1;
                    if (HH >= 24) {
                        HH = 0;
                        LCD_2digitafficheurbas(4,HH);
                        IncrementeDateUnJour(&Jour, &Mois, &Annee);
                        LCD_showBatteryLevel(niveau_batterie()); // mise  jour affichage batterie
                        ActiveDCF77(TRUE);
                    } else {
                        LCD_2digitafficheurbas(4,HH);
                    }
                } else {
                    LCD_2digitafficheurbas(2,MN);
                }
            } else {
                LCD_2digitafficheurbas(0,SS);
            }
            ITWatchdog = FALSE;
        }
        if (!(ITFront | ITWatchdog | ITMilliseconde)) LPM3; // On passe en basse consommation si plus d'IT  traiter
    }
}

